---@class zprite
---@field private _texture love.Image
---@field private _instance_vertex_count integer
---@field private _mesh love.Mesh
---@field private _instance_buffer love.Buffer
---@field private _instances {integer: {}[]}
---@field private _top_layer integer
---@field private _height_scale number
---@field private _changed boolean
---@field private _count integer
local zprite = {}
zprite.__index = zprite

local ZPRITE_SHADER = love.graphics.newShader([[
varying vec4 VsColorOut;

#ifdef VERTEX
layout (location = 3) attribute vec2 InstancePosition;
layout (location = 4) attribute vec2 InstanceTextureOffset;
layout (location = 5) attribute vec2 InstanceScale;
layout (location = 6) attribute float InstanceLayer;
layout (location = 7) attribute vec4 InstanceColor;

uniform float HeightScale;

vec4 position(mat4 transform_projection, vec4 vertex_position)
{
  vertex_position.xy *= InstanceScale;
  vertex_position.xy += InstancePosition.xy;
  vertex_position.z = InstanceLayer / 1024;
  VaryingTexCoord.xy *= InstanceScale;
  VaryingTexCoord.xy += InstanceTextureOffset;
  VsColorOut = InstanceColor;
  return (transform_projection * vertex_position) + vec4(0, (InstanceLayer * HeightScale) / love_ScreenSize.y, 0, 0);
}
#endif

#ifdef PIXEL
vec4 effect(vec4 color, Image tex, vec2 texture_coords, vec2 screen_coords)
{
  vec4 texture_color = Texel(tex, texture_coords);
  return texture_color * color * VsColorOut;
}
#endif
]])

local function ZPRITE_DEFAULT_COLORMAP(_)
	return 1, 1, 1, 1
end

---@param texture love.Image
---@param max_instances integer?
function zprite.new(texture, max_instances)
	if not texture then
		error("texture cannot be nil")
	end
	if max_instances and max_instances < 1 then
		error("max_instances cannot be less than 1")
	end

	local tw, th = texture:getPixelDimensions()
	local vertices = {
		{ 0, 0, 0, 0 },
		{ 1, 0, 1 / tw, 0 },
		{ 0, 1, 0, 1 / th },
		{ 1, 1, 1 / tw, 1 / th },
	}

	local obj = setmetatable({}, zprite)

	obj._texture = texture
	obj._instance_vertex_count = max_instances or 16384
	obj._mesh = love.graphics.newMesh({
		{ name = "VertexPosition", format = "floatvec2", location = 0 },
		{ name = "VertexTexCoord", format = "floatvec2", location = 1 },
	}, vertices, "strip", "static")
	obj._mesh:setTexture(texture)
	obj._instance_buffer = love.graphics.newBuffer({
		{ name = "InstancePosition", format = "floatvec2", location = 3 },
		{ name = "InstanceTextureOffset", format = "floatvec2", location = 4 },
		{ name = "InstanceScale", format = "floatvec2", location = 5 },
		{ name = "InstanceLayer", format = "float", location = 6 },
		{ name = "InstanceColor", format = "floatvec4", location = 7 },
	}, obj._instance_vertex_count, { vertex = true })
	obj._instances = {}
	obj._height_scale = 4
	obj._changed = false

	obj._mesh:attachAttribute(3, obj._instance_buffer, "perinstance")
	obj._mesh:attachAttribute(4, obj._instance_buffer, "perinstance")
	obj._mesh:attachAttribute(5, obj._instance_buffer, "perinstance")
	obj._mesh:attachAttribute(6, obj._instance_buffer, "perinstance")
	obj._mesh:attachAttribute(7, obj._instance_buffer, "perinstance")

	obj._top_layer = 0
	obj._count = 0

	return obj
end

--- Clear all drawn objects
function zprite:clear()
	-- no need to set _changed to true, as #self._instances will be 0
	self._instances = {}
	self._top_layer = 0
	self._count = 0
end

--- Add an 'entity' to be rendered
---@param x number
---@param y number
---@param quad love.Quad
---@param layer_count integer
---@param color_map function?
---@param base_layer integer?
function zprite:put(x, y, quad, layer_count, color_map, base_layer)
	if #self._instances + layer_count > self._instance_vertex_count then
		return
	end

	base_layer = base_layer or 0
	color_map = color_map or ZPRITE_DEFAULT_COLORMAP

	self._changed = true

	local qx, qy, qw, qh = quad:getViewport()
	local tw, th = self._texture:getPixelDimensions()

	for i = 1, layer_count do
		local r, g, b, a = color_map(i)
		if not a then
			a = 1
		end
		if not self._instances[i + base_layer] then
			self._instances[i + base_layer] = {}
			if i > self._top_layer then
				self._top_layer = i
			end
		end
		table.insert(self._instances[i], { x, y, qx / tw, qy / th, qw, qh, i + base_layer, r, g, b, a })
		self._count = self._count + 1
	end
end

--- Draw to the screen / current canvas
---@overload fun(self, x: number, y: number, r: number?, sx: number?, sy: number?, ox: number?, oy: number?, kx: number?, ky: number?)
---@overload fun(self, transform: love.Transform)
function zprite:draw(...)
	love.graphics.setShader(ZPRITE_SHADER)
	ZPRITE_SHADER:send("HeightScale", self._height_scale)

	if self._changed then
		local current = 1
		for i = 1, self._top_layer do
			if self._instances[i] then
				self._instance_buffer:setArrayData(self._instances[i], 1, current, #self._instances[i])
				current = current + #self._instances[i]
			end
		end
		self._changed = false
	end

	love.graphics.drawInstanced(self._mesh, self._count, ...)
end

return zprite
